library IEEE;
use IEEE.STD_LOGIC_1164.all;
--use IEEE.STD_LOGIC_ARITH.all;
--use IEEE.STD_LOGIC_SIGNED.all;
use IEEE.numeric_std.all;

-------------------------------------------------------------------------------
-- entity
-------------------------------------------------------------------------------

entity de2_audio is
port (
    clk : in std_logic;                            
    reset_n : in std_logic;

    -- Bus master signals
    address      : in std_logic_vector (7 downto 0);
    byteenable   : in std_logic_vector (1 downto 0);
    writedata    : in std_logic_vector (15 downto 0);
    read         : in std_logic;
    write        : in std_logic;
    chipselect   : in std_logic;

     -- Slave signals
    readdata     : out std_logic_vector (15 downto 0);
    waitrequest  : out std_logic;   
   
    -- Audio interface signals
    AUD_CLK         : in std_logic;                  --    18.43MHz audio clock AUD_XCK
    AUD_ADCLRCK  : out std_logic;                    --    Audio CODEC ADC LR Clock
    AUD_ADCDAT   : in  std_logic;                    --    Audio CODEC ADC Data
    AUD_DACLRCK  : out std_logic;                    --    Audio CODEC DAC LR Clock
    AUD_DACDAT   : out std_logic;                    --    Audio CODEC DAC Data
    AUD_BCLK     : inout std_logic;                  --    Audio CODEC Bit-Stream Clock
    AUD_XCK      : out std_logic;
    -- Test signals --
    ledr    : out std_logic_vector (17 downto 0)
  );
end de2_audio;

-------------------------------------------------------------------------------
-- architecture
-------------------------------------------------------------------------------

architecture imp of de2_audio is

      component de2_fifo
      port (
            data        : IN STD_LOGIC_VECTOR (15 DOWNTO 0);
            rdclk        : IN STD_LOGIC ;
            rdreq        : IN STD_LOGIC ;
            wrclk        : IN STD_LOGIC ;
            wrreq        : IN STD_LOGIC ;
            q            : OUT STD_LOGIC_VECTOR (15 DOWNTO 0);
            rdempty        : OUT STD_LOGIC ;
            wrfull        : OUT STD_LOGIC ;
            wrusedw        : OUT STD_LOGIC_VECTOR (12 DOWNTO 0)
      );
      end component;

    component de2_wm8731_audio_in is
    port (
        clk : in std_logic;       --  Audio CODEC Chip Clock AUD_XCK (18.43 MHz)
        reset_n : in std_logic;
        data_out : out std_logic_vector(15 downto 0);
        audio_req : out std_logic;

        -- Audio interface signals
        AUD_ADCLRCK  : out std_logic;   --    Audio CODEC ADC LR Clock
        AUD_ADCDAT   : in  std_logic;   --    Audio CODEC ADC Data
        AUD_BCLK     : inout std_logic  --    Audio CODEC Bit-Stream Clock
    );
    end component;
   
    component de2_wm8731_audio_out is
    port (
        clk : in std_logic;       --  Audio CODEC Chip Clock AUD_XCK (18.43 MHz)
        reset_n : in std_logic;
        test_mode : in std_logic;       --    Audio CODEC controller test mode
        data_in : in std_logic_vector(15 downto 0);   
        audio_req : out std_logic;
     
        -- Audio interface signals
        AUD_DACLRCK  : out std_logic;   --    Audio CODEC DAC LR Clock
        AUD_DACDAT   : out std_logic   --    Audio CODEC DAC Data
    );
    end component;   

    signal reset : std_logic;

    signal dac_request : std_logic;
    signal adc_request : std_logic;
    signal data_from_bus : std_logic_vector (15 downto 0);
    signal data_to_bus : std_logic_vector (15 downto 0);
    signal adc_data : std_logic_vector (15 downto 0);
    signal dac_data : std_logic_vector (15 downto 0);
	signal adc_rdempty : std_logic;
    signal adc_wrfull : std_logic;
    signal dac_rdempty : std_logic;
    signal dac_wrfull : std_logic;
    signal writefifo   : std_logic;
    signal readfifo       : std_logic;   
    signal adc_rdreq   : std_logic;   
    signal dac_wrreq   : std_logic;   
   
    signal adc_used_buf : std_logic_vector (12 downto 0);
    signal dac_used_buf : std_logic_vector (12 downto 0);

    signal adc_stop : std_logic;
    signal dac_stop : std_logic;

    --test
    signal ledr_count : std_logic_vector (16 downto 0);
    signal led_ctrl : std_logic;
   
    --Audio clk dividers
    signal sampling_clk : std_logic;
    signal audio_counter : unsigned (13 downto 0);
    signal counter_aud_adclkrck : unsigned (15 downto 0);
    signal LRCK: std_logic;
    signal bus_state: std_logic_vector (2 downto 0);

	type state_t is (a0, a1, a2, d0, d1, d2);
	signal adc_state, adc_new_state, dac_state, dac_new_state : state_t;
   
begin

    adc_rdreq <= readfifo;
    dac_wrreq <= writefifo;
	
	-- waitrequest is combinational
	waitrequest <= (chipselect and adc_rdempty and read) or (chipselect and dac_wrfull and write);
   
	SM_ADC: process(adc_state, adc_rdempty, read, adc_stop)
	begin
		case adc_state is
			when a0 =>
				if (adc_rdempty = '0' and read = '1') then
					adc_new_state <= a1;
				else
					adc_new_state <= a0;
				end if;					
			when a1 =>
				if (adc_stop = '1') then
					adc_new_state <= a2;
				else
					adc_new_state <= a1;
				end if;				
			when a2 =>
				if (read = '1' and adc_rdempty = '0') then
					adc_new_state <= a1;
				else
					adc_new_state <= a0;
				end if;				
			when others =>
				adc_new_state <= a0;
		end case;
	end process SM_ADC;
	
	SM_DAC: process(dac_state, dac_wrfull, write, dac_stop)
	begin
		case dac_state is
			when d0 =>
				if (dac_wrfull = '0' and write = '1') then
					dac_new_state <= d1;
				else
					dac_new_state <= d0;
				end if;					
			when d1 =>
				if (dac_stop = '1') then
					dac_new_state <= d2;
				else
					dac_new_state <= d1;
				end if;
				
			when d2 =>
				if (write = '1' and dac_wrfull = '0') then
					dac_new_state <= d1;
				else 
					dac_new_state <= d0;
				end if;					
			when others =>
				dac_new_state <= d0;
		end case;
	end process SM_DAC;	
	
	ST_ADC: process(clk, reset_n)
	begin
		if reset_n = '0' then
			adc_state <= a0;
		elsif rising_edge(clk) then
			if chipselect = '1' then
				adc_state <= adc_new_state;
			end if;
		end if;
	end process ST_ADC;
	
	ST_DAC: process(clk, reset_n)
	begin
		if reset_n = '0' then
			dac_state <= d0;
		elsif rising_edge(clk) then
			if chipselect = '1' then
				dac_state <= dac_new_state;
			end if;
		end if;
	end process ST_DAC;
	
	BusComm: process(clk, reset_n)
	begin
		if reset_n = '0' then
			readdata <= (others => '0');
            --waitrequest <= '0';	
			adc_stop <= '0';
			dac_stop <= '0';
			readfifo <= '0';
			writefifo <= '0';
		elsif rising_edge(clk) then
			if adc_new_state = a0 then
				if chipselect = '1' then
					readfifo <= '0';
					--waitrequest <= '0'; 
					adc_stop <= '0';
				end if;
			elsif adc_new_state = a1 then
				if chipselect = '1' then
					--if (address(0) = '1' and read = '1') then
					if (read = '1') then
						readdata <= data_to_bus;
						readfifo <= '1';
						--waitrequest <= '1';
						adc_stop <= '1';
					end if;
                end if;
			elsif adc_new_state = a2 then
				if chipselect = '1' then
					--waitrequest <= '0';
					adc_stop <= '0';
					readfifo <= '0';
				end if;
			end if;
			
			if dac_new_state = d0 then
				if chipselect = '1' then
					writefifo <= '0';
					--waitrequest <= '0';
					dac_stop <= '0';
				end if;
			elsif dac_new_state = d1 then
				if chipselect = '1' then
					--if (address(0) = '1' and write = '1') then
					if (write = '1') then
						data_from_bus <= writedata;   
						writefifo <= '1';
						--waitrequest <= '1';   
						dac_stop <= '1';
					end if;
				end if;
			elsif dac_new_state = d2 then
				if chipselect = '1' then
					--waitrequest <= '0';
					dac_stop <= '0';
					writefifo <= '0';	
				end if;			
			end if;
			
		end if;
	end process BusComm;
     
    process(AUD_BCLK)
    begin
    if rising_edge(AUD_BCLK) then
        counter_aud_adclkrck <= counter_aud_adclkrck + 1;
    end if;

    if (counter_aud_adclkrck = "0000000000000000") then
        led_ctrl <= not led_ctrl;
    end if;
    end process;
   
    AUD_XCK <= AUD_CLK;

    ADC : de2_wm8731_audio_in
    port map (
        clk => AUD_CLK,
        reset_n => reset_n,
        data_out => adc_data,
        audio_req => adc_request,

        AUD_ADCLRCK => AUD_ADCLRCK,
        AUD_ADCDAT => AUD_ADCDAT,
        AUD_BCLK => AUD_BCLK
    );
   
    DAC : de2_wm8731_audio_out
    port map (
        clk => AUD_CLK,
        reset_n => reset_n,     
        test_mode => '0',
        audio_req => dac_request,
        data_in => dac_data,         

        AUD_DACLRCK  => AUD_DACLRCK,
        AUD_DACDAT   => AUD_DACDAT
    ); 

    ADC_FIFO : de2_fifo
    port map (
        data        => adc_data,
        rdclk        => clk,
        rdreq        => adc_rdreq,
        wrclk        => AUD_CLK,
        wrreq        => adc_request,
        q            => data_to_bus,
        rdempty        => adc_rdempty,
        wrfull        => adc_wrfull,
        wrusedw        => adc_used_buf   
    );

    DAC_FIFO : de2_fifo
    port map (
        data        => data_from_bus,
        rdclk        => AUD_CLK,
        rdreq        => dac_request,
        wrclk        => clk,
        wrreq        => dac_wrreq,
        q            => dac_data,
        rdempty        => dac_rdempty,
        wrfull        => dac_wrfull,
        wrusedw        => dac_used_buf   
    );   
   
      ledr(1)  <= led_ctrl;
      ledr(2)  <= dac_wrfull;
      ledr(3)  <= adc_rdempty;
      ledr(0)  <= '0';

end architecture imp;