/* wave.c
 *
 * Data input functions.
 */

#include "types.h"

#define ISIZE samp_per_frame*32 /* input buffer size in longs */

static unsigned long *ibuff;  /* input buffer pointer */
static int
  ihead,             /* write pointer */
  itail;             /* read pointer */

/*
 * wave_close:
 * -----------
 */
void wave_close(void)
{
  fclose(config.wave.file);
}

/*
 * wave_open:
 * ----------
 * Opens and verifies the header of the Input Wave file. The file pointer is
 * left pointing to the start of the samples.
 */
void wave_open(int raw, int mono_from_stereo)
{

  static char *channel_mappings[] = {NULL,"mono","stereo"};
  static char *file_type[] = {"RIFF-WAVE PCM","RAW DATA PCM"};

  /* Wave file headers can vary from this, but we're only intereseted in this format */
  struct wave_header
  {
    char riff[4];             /* "RIFF" */
    unsigned long size;       /* length of rest of file = size of rest of header(36) + data length */
    char wave[4];             /* "WAVE" */
    char fmt[4];              /* "fmt " */
    unsigned long fmt_len;    /* length of rest of fmt chunk = 16 */
    unsigned short tag;       /* MS PCM = 1 */
    unsigned short channels;  /* channels, mono = 1, stereo = 2 */
    unsigned long samp_rate;  /* samples per second = 44100 */
    unsigned long byte_rate;  /* bytes per second = samp_rate * byte_samp = 176400 */
    unsigned short byte_samp; /* block align (bytes per sample) = channels * bits_per_sample / 8 = 4 */
    unsigned short bit_samp;  /* bits per sample = 16 for MS PCM (format specific) */
    char data[4];             /* "data" */
    unsigned long length;     /* data length (bytes) */
  } header;

  /* open input file */
  if((config.wave.file = fopen(config.infile,"rb")) == NULL)
    error("Unable to open file");
  if((ibuff = (unsigned long *)malloc(ISIZE*sizeof(long))) == NULL)
    error("Unable to allocate input buffer");
  ihead = itail = 0;

  if(raw)
  {

      fseek(config.wave.file, 0, SEEK_END); // find length of file
      header.length = ftell(config.wave.file);
      fseek(config.wave.file, 0, SEEK_SET);

      header.channels = 1;
      header.samp_rate = 8000; /* 44100 if not set by user */
      header.bit_samp = 16;     /* assumed */
      header.byte_samp = header.channels * header.bit_samp / 8;
      header.byte_rate = header.samp_rate * header.byte_samp;

  }
  else // wave
  {
      if (fread(&header, sizeof(header), 1, config.wave.file) != 1)
        error("Invalid Header");
      if(strncmp(header.riff,"RIFF",4) != 0) error("Not a MS-RIFF file");
      if(strncmp(header.wave,"WAVE",4) != 0) error("Not a WAVE audio");
      if(strncmp(header.fmt, "fmt ",4) != 0) error("Can't find format chunk");
      if(header.tag != 1)                    error("Unknown WAVE format");
      if(header.channels > 2)                error("More than 2 channels");
      if(header.bit_samp != 16)              error("Not 16 bit");
      if(strncmp(header.data,"data",4) != 0) error("Can't find data chunk");
  }

  config.wave.channels      = header.channels;
  config.wave.samplerate    = header.samp_rate;
  config.wave.bits          = header.bit_samp;
  config.wave.total_samples = header.length / header.byte_samp;
  config.wave.length        = header.length / header.byte_rate;

  if((config.wave.channels == 1)) //|| mono_from_stereo)
  {
    config.mpeg.channels = 1;
    config.mpeg.mode = MODE_MONO;
  }

  printf("%s, %s %ldHz %dbit\n",// Length: %2ld:%2ld:%2ld\n",
          file_type[1],
          channel_mappings[config.wave.channels],
          config.wave.samplerate,
          config.wave.bits);
          //config.wave.length/3600,(config.wave.length/60)%60,config.wave.length%60);
}

/*
 * wave_get:
 * ---------
 * Reads samples from the file in nice large chunks of longs. A long can
 * hold one stereo or two mono samples. The buffer holds a multiple of frames.
 * Returns the address of the start of the next frame or NULL if there are no
 * more. The last frame will be padded with zero's.
 *
 * note, This doesn't speed things up much, the filesystem has it's own buffers.
 * But it does reduce disc access.
 */
unsigned long int *wave_get(void)
{
  static int my_eof;
  int i,m,n,p;
  unsigned long *buff;

  n = config.mpeg.samples_per_frame >> (2 - config.wave.channels);

  m = ihead - itail; /* available in buffer */

  if(m < n) /* require more than available */
  {
    if(my_eof) /* must be the last frame */
    {
      if(!m)
        return 0;

      buff = &ibuff[itail];
      itail += m;
      return buff;
    }

    /* fill buffer from file */
    p = fread(ibuff,sizeof(long),ISIZE,config.wave.file);
    ihead = p;

    if(p < ISIZE) /* not enough left to fill buffer */
    {
      my_eof = 1;
      for(i=p; i<ISIZE; i++) /* pad */
        ibuff[i] = 0;
    }

    if(p > n) /* more than required */
      p = n;

    itail = p;
    return ibuff;
  }
  else /* more than required */
  {
    buff = &ibuff[itail];
    itail += n;
    return buff;
  }
}

